@@ -1,9 +1,9 @@
 #!/usr/bin/env python
-"""Lister 1.5.1
+"""Lister 1.6
 Pretty-printer for SimPy class objects
-'$Revision: 1.1.1.2.4.1 $ $Date: 2005/01/30 11:41:26 $ kgm'
+'$Revision: 1.1.1.2.4.1.4.1 $ $Date: 2005/05/27 05:16:50 $ kgm'
 """
-__version__ = '1.5.1 February 2005'
+__version__ = '1.6 June 2005'
 class Lister(object):
     indent=0
     def __str__(self):
