@@ -1,8 +1,8 @@
 #!/usr/bin/env python
-"""Monitor 1.5.1
- '$Revision: 1.1.1.4.4.1 $ $Date: 2005/01/30 11:41:26 $ kgm'
+"""Monitor 1.6
+ '$Revision: 1.1.1.4.4.1.4.1 $ $Date: 2005/05/27 05:17:50 $ kgm'
 This dummy module is only provided for backward compatibility.
 It does nothing. class Monitor is now part of Simulation
 and SimulationXXX."""
-__version__ ='1.5.1 February 2005'
+__version__ ='1.6 June 2005'
 pass   
