@@ -1,7 +1,7 @@
 #!/usr/bin/env python
-""" SimPlot 1.5.1
+""" SimPlot 1.6
 
-$Revision: 1.1.1.1.4.2 $ $Date: 2005/01/30 11:41:26 $ kgm
+$Revision: 1.1.1.1.4.2.4.1 $ $Date: 2005/05/27 05:17:13 $ kgm
 
 LICENSE:
 Copyright (C) 2002, 2005  Klaus G. Muller, Tony Vignaux
@@ -38,7 +38,7 @@ Change history:
     Feb 2004:       Release with SimPy 1.4
 
 """
-__version__= '1.5.1 February 2005'
+__version__= '1.6 June 2005'
 from Tkinter import *
 from Canvas import Line, CanvasText, Rectangle
 from tkMessageBox import *
