@@ -1,6 +1,6 @@
 #!/usr/bin/env python
-"""Simulation 1.5.1
-__version__ = '$Revision: 1.1.1.4.4.2 $ $Date: 2005/01/30 11:41:26 $ kgm'
+"""Simulation 1.6
+__version__ = '$Revision: 1.1.1.4.4.2.4.14 $ $Date: 2005/06/09 09:12:27 $ kgm'
 LICENSE:
 Copyright (C) 2002,2005  Klaus G. Muller, Tony Vignaux
 mailto: kgmuller@xs4all.nl and Tony.Vignaux@vuw.ac.nz
@@ -209,15 +209,29 @@ Based on generators (Python 2.2 and later)
     12 Jan 2005: SimPy version 1.5.1
                  Changes in this module: Monitor objects now have a default name
                                          'a_Monitor'
+                                         
+    29 Mar 2005: Start SimPy 1.6: compound "yield request" statements
+    
+    05 Jun 2005: Fixed bug in _request method -- waitMon did not work properly in
+                 preemption case
+                 
+    09 Jun 2005: Added test in 'activate' to see whether 'initialize()' was called first.
+                                         
   
 """
 from __future__ import generators
 from SimPy.Lister import *
 import bisect
 import types
+import sys
 __TESTING=False
-__version__="1.5.1 February 2005"
-if __TESTING: print "SimPy.Simulation %s" %__version__
+__version__="1.6 June 2005"
+if __TESTING: 
+    print "SimPy.Simulation %s" %__version__,
+    if __debug__:
+        print "__debug__ on"
+    else:
+        print
 
 hold=1234
 passivate=5678
@@ -266,6 +280,11 @@ class Simerror(Exception):
 
     def __str__(self):
         return `self.value`
+        
+class FatalSimerror(Simerror):
+    def __init__(self,value):
+        Simerror.__init__(self,value)
+        self.value=value
     
 class Process(Lister):
     """Superclass of classes which may use generator functions"""
@@ -335,6 +354,21 @@ class Process(Lister):
         """
         self._inInterrupt= False
 
+    def acquired(self,res):
+        """Tests whether resource res was acquired when proces reactivated.
+        If yes, the parallel wakeup process is killed.
+        If not, process is removed from res' waitQ (reneging).
+        """
+        test=self in res.activeQ
+        
+        if test:
+            self.cancel(self._holder)
+        else:
+            res.waitQ.remove(self)
+            if res.monitored:
+                res.waitMon.observe(len(res.waitQ),t=now())
+        return test        
+
 def allEventNotices():
     """Returns string with eventlist as list of tuples (eventtime,action)"""
     return `[(_e.events[i][0],[inst.who.name for inst in _e.events[1]]) \
@@ -346,7 +380,7 @@ class __Evlist:
         #has the structure {<time1>:(ev notice1, ev notice2, . . ),
 #                           <time2>:(ev notice 3,...),..}
         self.events={}
-        #always sorted list of event times (keys for self.events)	
+        #always sorted list of event times (keys for self.events)    
         self.timestamps=[]
 
     def _post(self,what,at,prior=False):
@@ -380,11 +414,11 @@ class __Evlist:
                     self.events[thistime].remove(n)
                     whom._nextTime = None 
                     if not self.events[thistime]:
-            # no other ev notices for thistime
+                        # no other ev notices for thistime
                         del(self.events[thistime])
                         item_delete_point = bisect.bisect(self.timestamps,
                                               thistime)
-                        del self.timestamps[item_delete_point-1]	          
+                        del self.timestamps[item_delete_point-1]              
                                               
     def _nextev(self):
         """Retrieve next event from event list"""
@@ -419,10 +453,12 @@ class _Action:
         self.process=generator
 
     def __str__(self):
-        return "Action%s" %((self.who,self.process))
+        return "Action %s %s" %((self.who.name,self.process))
 
 def activate(object,process,at="undefined",delay="undefined",prior=False):
     """Application function to activate passive process.""" 
+    if _e is None:
+        raise Simerror("Fatal SimPy error: simulation is not initialized (call initialize() first)")
     if not (type(process) == types.GeneratorType):
         raise Simerror("Fatal SimPy error: activating function which"+
                        " is not a generator (contains no 'yield')")
@@ -556,7 +592,10 @@ class Resource(Lister):
                 # remove from activeQ
                 self.activeQ.remove(z)
                 # put into front of waitQ
-                self.waitQ.insert(0,z)
+                self.waitQ.insert(0,z)               
+                # if self is monitored, update waitQ monitor
+                if self.monitored:
+                    self.waitMon.observe(len(self.waitQ),now())                    
                 # record that it has been preempted
                 z._preempted = 1
                 # passivate re-queued process
@@ -776,7 +815,56 @@ def holdfunc(a):
     a[0][1]._hold(a)
 
 def requestfunc(a):
-    a[0][2]._request(a)
+    """Handles 'yield request,self,res' and 'yield (request,self,res),(<code>,self,<par>)'.
+    """
+    if type(a[0][0])==tuple:
+        ## Compound yield request statement
+        ## first tuple in ((request,self,res),(xx,self,yy))
+        b=a[0][0]
+        ## b[2]==res (the resource requested)
+        ##process the first part of the compound yield statement
+        ##a[1] is the _Action instance (a[1].who==process owner, a[1].process==generator)
+        b[2]._request(arg=(b,a[1]))
+        ##deal with add-on condition to command
+        ##Trigger processes for reneging
+        class _Holder(Process):
+            """Provides timeout process"""
+            def trigger(self,delay):
+                yield hold,self,delay
+                if not proc in b[2].activeQ:
+                    reactivate(proc)
+
+        class _EventWait(Process):
+            """Provides event waiting process"""
+            def trigger(self,event):
+                yield waitevent,self,event
+                if not proc in b[2].activeQ:
+                    reactivate(proc)
+               
+        #activate it
+        proc=a[0][0][1] # the process to be woken up
+        actCode=a[0][1][0]
+        if actCode==hold:
+            proc._holder=_Holder()
+            ##                                          the timeout delay
+            activate(proc._holder,proc._holder.trigger(a[0][1][2]))
+        elif actCode==waituntil:
+            raise FatalSimerror("illegal code for reneging: waituntil")
+            #~ proc._holder=_CondWait()
+        elif actCode==waitevent:
+            proc._holder=_EventWait(proc.name)
+            ##                                          the event
+            activate(proc._holder,proc._holder.trigger(a[0][1][2]))
+        elif actCode==queueevent:
+            raise FatalSimerror("illegal code for reneging: queueevent")
+            #~ proc._holder=_QueueEvent(proc.name)
+        else:
+            raise FatalSimerror("illegal code for reneging %s"%actCode)
+        # process the first part of the yield stmnt
+        #b[2]._request((b,a[1]))
+    else:
+        ## Simple yield request command
+        a[0][2]._request(a)
 
 def releasefunc(a):
     a[0][2]._release(a)
@@ -824,14 +912,25 @@ def simulate(until=0):
     yield release,self,<Resource> : release 1 unit to <Resource>
 
     yield waitevent,self,<SimEvent>|[<Evt1>,<Evt2>,<Evt3), . . . ]:
-    	wait for one or more of several events
-    	
+        wait for one or more of several events
+        
 
     yield queueevent,self,<SimEvent>|[<Evt1>,<Evt2>,<Evt3), . . . ]:
-    	queue for one or more of several events
+        queue for one or more of several events
     
     yield waituntil,self,cond : wait for arbitrary condition    
 
+    EXTENSIONS:
+    Request with timeout reneging:
+    yield (request,self,<Resource>),(hold,self,<patience>) :
+        request 1 unit from <Resource>. If unit not acquired before time period
+        <patience> has passed, self leaves waitQ (reneges).
+
+    Request with event-based reneging:
+    yield (request,self,<Resource>),(waitevent,self,<event>):
+        request 1 unit from <Resource>. If <event> occurs before unit
+        acquired, leave waitQ (renege).
+
     Event notices get posted in event-list by scheduler after "yield" or by 
     "activate"/"reactivate" functions.
     
@@ -850,13 +949,27 @@ def simulate(until=0):
     dispatch={hold:holdfunc,request:requestfunc,release:releasefunc,
               passivate:passivatefunc,waitevent:waitevfunc,queueevent:queueevfunc,
               waituntil:waituntilfunc}
+    commandcodes=dispatch.keys()
+    commandwords={hold:"hold",request:"request",release:"release",passivate:"passivate",
+        waitevent:"waitevent",queueevent:"queueevent",waituntil:"waituntil"}
     while not _stop and _t<=_endtime:
         try:
-            a=_e._nextev()        
+            a=_e._nextev()          
             if not a[0]==None:
-                command = a[0][0]
+                ## 'a' is tuple "(<yield command>, <action>)"  
+                if type(a[0][0])==tuple:
+                    ##allowing for yield (request,self,res),(waituntil,self,cond)
+                    command=a[0][0][0]
+                else: 
+                    command = a[0][0]
+                if __debug__:
+                    if not command in commandcodes:
+                        raise FatalSimerror("Fatal error: illegal command: yield %s"%command)
                 dispatch[command](a)     
-        except Simerror, error:
+        except FatalSimerror,error:
+            print "SimPy: "+error.value
+            sys.exit(1)
+        except Simerror,error:
             message="SimPy: "+error.value
             _stop = True
         if _wustep:
@@ -1231,84 +1344,205 @@ def testwaituntil():
     print
     print simulate(until=9*60)
     
+## -------------------------------------------------------------
+##                    TEST COMPOUND "YIELD REQUEST" COMMANDS
+## -------------------------------------------------------------
+
+## -------------------------------------------------------------
+##             TEST "yield (request,self,res),(hold,self,delay)"
+##                   == timeout renege
+## -------------------------------------------------------------
+
+class JobTO(Process):
+   """ Job class for testing timeout reneging
+   """
+   def __init__(self,server=None,name=""):
+        Process.__init__(self,name)
+        self.res=server
+        self.gotResource=None
+        
+   def execute(self,timeout,usetime):       
+        yield (request,self,self.res),(hold,self,timeout)
+        if self.acquired(self.res):
+            self.gotResource=True
+            yield hold,self,usetime
+            yield release,self,self.res
+        else:
+            self.gotResource=False
+            
+def testNoTimeout():
+    """Test that resource gets acquired without timeout
+    """
+    res=Resource(name="Server",capacity=1)
+    initialize()
+    usetime=5
+    timeout=1000000
+    j1=JobTO(server=res,name="Job_1")
+    activate(j1,j1.execute(timeout=timeout,usetime=usetime))
+    j2=JobTO(server=res,name="Job_2")
+    activate(j2,j2.execute(timeout=timeout,usetime=usetime))      
+    simulate(until=2*usetime)
+    assert now()==2*usetime,"time not ==2*usetime" 
+    assert j1.gotResource and j2.gotResource,\
+        "at least one job failed to get resource"
+    assert not (res.waitQ or res.activeQ),\
+        "job waiting or using resource"
+            
+def testTimeout1():
+    """Test that timeout occurs when resource busy
+    """
+    res=Resource(name="Server",capacity=1,monitored=True)
+    initialize()
+    usetime=5
+    timeout=3
+    j1=JobTO(server=res,name="Job_1")
+    activate(j1,j1.execute(timeout=timeout,usetime=usetime))
+    j2=JobTO(server=res,name="Job_2")
+    activate(j2,j2.execute(timeout=timeout,usetime=usetime))      
+    simulate(until=2*usetime)
+    assert(now()==usetime),"time not ==usetime"
+    assert(j1.gotResource),"Job_1 did not get resource"
+    assert(not j2.gotResource),"Job_2 did not renege"
+    assert not (res.waitQ or res.activeQ),\
+        "job waiting or using resource"
+
+def testTimeout2():
+    """Test that timeout occurs when resource has no capacity free
+    """
+    res=Resource(name="Server",capacity=0)
+    initialize()
+    usetime=5
+    timeout=3
+    j1=JobTO(server=res,name="Job_1")
+    activate(j1,j1.execute(timeout=timeout,usetime=usetime))
+    j2=JobTO(server=res,name="Job_2")
+    activate(j2,j2.execute(timeout=timeout,usetime=usetime))      
+    simulate(until=2*usetime)
+    assert now()==timeout,"time %s not == timeout"%now()
+    assert not j1.gotResource,"Job_1 got resource"
+    assert not j2.gotResource,"Job_2 got resource"
+    assert not (res.waitQ or res.activeQ),\
+        "job waiting or using resource"  
+
+## ------------------------------------------------------------------
+##             TEST "yield (request,self,res),(waitevent,self,event)"
+##                   == event renege
+## ------------------------------------------------------------------
+class JobEvt(Process):
+   """ Job class for testing event reneging
+   """
+   def __init__(self,server=None,name=""):
+        Process.__init__(self,name)
+        self.res=server
+        self.gotResource=None
+        
+   def execute(self,event,usetime):       
+        yield (request,self,self.res),(waitevent,self,event)
+        if self.acquired(self.res):
+            self.gotResource=True
+            yield hold,self,usetime
+            yield release,self,self.res
+        else:
+            self.gotResource=False
+            
+class JobEvtMulti(Process):
+   """ Job class for testing event reneging with multi-event lists
+   """
+   def __init__(self,server=None,name=""):
+        Process.__init__(self,name)
+        self.res=server
+        self.gotResource=None
+        
+   def execute(self,eventlist,usetime):       
+        yield (request,self,self.res),(waitevent,self,eventlist)
+        if self.acquired(self.res):
+            self.gotResource=True
+            yield hold,self,usetime
+            yield release,self,self.res
+        else:
+            self.gotResource=False
+            
+class FireEvent(Process):
+    """Fires reneging event
+    """
+    def fire(self,fireDelay,event):
+        yield hold,self,fireDelay
+        event.signal()
+            
+def testNoEvent():
+    """Test that processes acquire resource normally if no event fires
+    """
+    res=Resource(name="Server",capacity=1)
+    event=SimEvent("Renege_trigger") #never gets fired
+    initialize()
+    usetime=5
+    j1=JobEvt(server=res,name="Job_1")
+    activate(j1,j1.execute(event=event,usetime=usetime))
+    j2=JobEvt(server=res,name="Job_2")
+    activate(j2,j2.execute(event=event,usetime=usetime))      
+    simulate(until=2*usetime)
+    # Both jobs should get server (in sequence)
+    assert now()==2*usetime,"time not ==2*usetime" 
+    assert j1.gotResource and j2.gotResource,\
+        "at least one job failed to get resource"
+    assert not (res.waitQ or res.activeQ),\
+        "job waiting or using resource"
+ 
+def testWaitEvent1():
+    """Test that signalled event leads to renege when resource busy
+    """
+    res=Resource(name="Server",capacity=1)
+    initialize()
+    event=SimEvent("Renege_trigger")
+    usetime=5
+    eventtime=1
+    j1=JobEvt(server=res,name="Job_1")
+    activate(j1,j1.execute(event=event,usetime=usetime))
+    j2=JobEvt(server=res,name="Job_2")
+    activate(j2,j2.execute(event=event,usetime=usetime))
+    f=FireEvent(name="FireEvent")
+    activate(f,f.fire(fireDelay=eventtime,event=event))           
+    simulate(until=2*usetime)
+    # Job_1 should get server, Job_2 renege
+    assert(now()==usetime),"time not ==usetime"
+    assert(j1.gotResource),"Job_1 did not get resource"
+    assert(not j2.gotResource),"Job_2 did not renege"
+    assert not (res.waitQ or res.activeQ),\
+        "job waiting or using resource"
+            
+def testWaitEvent2():
+    """Test that renege-triggering event can be one of an event list
+    """
+    res=Resource(name="Server",capacity=1)
+    initialize()
+    event1=SimEvent("Renege_trigger_1")
+    event2=SimEvent("Renege_trigger_2")
+    usetime=5
+    eventtime=1 #for both events
+    j1=JobEvtMulti(server=res,name="Job_1")
+    activate(j1,j1.execute(eventlist=[event1,event2],usetime=usetime))
+    j2=JobEvtMulti(server=res,name="Job_2")
+    activate(j2,j2.execute(eventlist=[event1,event2],usetime=usetime))
+    f1=FireEvent(name="FireEvent_1")
+    activate(f1,f1.fire(fireDelay=eventtime,event=event1))
+    f2=FireEvent(name="FireEvent_2")
+    activate(f2,f2.fire(fireDelay=eventtime,event=event2))             
+    simulate(until=2*usetime)
+    # Job_1 should get server, Job_2 should renege
+    assert(now()==usetime),"time not ==usetime"
+    assert(j1.gotResource),"Job_1 did not get resource"
+    assert(not j2.gotResource),"Job_2 did not renege"
+    assert not (res.waitQ or res.activeQ),\
+        "job waiting or using resource"
+
 if __name__ == "__main__":
     print "SimPy.Simulation %s" %__version__
+    testNoTimeout()
+    testTimeout1()
+    testTimeout2()
+    testNoEvent()
+    testWaitEvent1()
+    testWaitEvent2()
     test_demo()
     test_interrupt()
-    testSimEvents()
     testwaituntil()
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
