@@ -1,6 +1,6 @@
 #!/usr/bin/env python
-"""SimulationRT 1.5.1
-__version__ = '$Revision: 1.1.1.1.4.2 $ $Date: 2005/01/30 11:41:26 $ kgm'
+"""SimulationTrace 1.6
+__version__ = '$Revision: 1.1.1.1.4.2.4.6 $ $Date: 2005/06/09 09:13:14 $ kgm'
 LICENSE:
 Copyright (C) 2002, 2005  Klaus G. Muller, Tony Vignaux
 mailto: kgmuller@xs4all.nl and Tony.Vignaux@vuw.ac.nz
@@ -55,6 +55,14 @@ Based on generators (Python 2.2 and later)
     12 Jan 2005: SimPy version 1.5.1
                  Changes in this module: Monitor objects now have a default name
                                          'a_Monitor'
+                                         
+    29 Mar 2005: Start SimPy 1.6: compound "yield request" statements
+    
+    05 Jun 2005: Fixed bug in _request method -- waitMon did not work properly in
+                 preemption case
+                 
+    09 Jun 2005: Added test in 'activate' to see whether 'initialize()' was called first.
+                                         
   
 """
 from __future__ import generators
@@ -64,8 +72,14 @@ import types
 import time
 import sys
 __TESTING=False
-__version__="1.5.1 February 2005"
-if __TESTING: print "SimPy.SimulationRT %s" %__version__
+__version__="1.6 June 2005"
+if __TESTING: 
+    print "SimPy.SimulationRT %s" %__version__,
+    if __debug__:
+        print "__debug__ on"
+    else:
+        print
+
 hold=1234
 passivate=5678
 request=135
@@ -121,6 +135,11 @@ class Simerror(Exception):
 
     def __str__(self):
         return `self.value`
+        
+class FatalSimerror(Simerror):
+    def __init__(self,value):
+        Simerror.__init__(self,value)
+        self.value=value
     
 class Process(Lister):
     """Superclass of classes which may use generator functions"""
@@ -190,6 +209,21 @@ class Process(Lister):
         """
         self._inInterrupt= False
 
+    def acquired(self,res):
+        """Tests whether resource res was acquired when proces reactivated.
+        If yes, the parallel wakeup process is killed.
+        If not, process is removed from res' waitQ (reneging).
+        """
+        test=self in res.activeQ
+        
+        if test:
+            self.cancel(self._holder)
+        else:
+            res.waitQ.remove(self)
+            if res.monitored:
+                res.waitMon.observe(len(res.waitQ),t=now())
+        return test        
+
 def allEventNotices():
     """Returns string with eventlist as list of tuples (eventtime,action)"""
     return `[(_e.events[i][0],[inst.who.name for inst in _e.events[1]]) \
@@ -201,7 +235,7 @@ class __Evlist:
         #has the structure {<time1>:(ev notice1, ev notice2, . . ),
 #                           <time2>:(ev notice 3,...),..}
         self.events={}
-        #always sorted list of event times (keys for self.events)	
+        #always sorted list of event times (keys for self.events)    
         self.timestamps=[]
         self.real_time=False
         self.rel_speed=1     
@@ -237,11 +271,11 @@ class __Evlist:
                     self.events[thistime].remove(n)
                     whom._nextTime = None 
                     if not self.events[thistime]:
-            # no other ev notices for thistime
+                        # no other ev notices for thistime
                         del(self.events[thistime])
                         item_delete_point = bisect.bisect(self.timestamps,
                                               thistime)
-                        del self.timestamps[item_delete_point-1]	          
+                        del self.timestamps[item_delete_point-1]              
                                               
     def _nextev(self):
         """Retrieve next event from event list"""
@@ -279,10 +313,12 @@ class _Action:
         self.process=generator
 
     def __str__(self):
-        return "Action%s" %((self.who,self.process))
+        return "Action %s %s" %((self.who.name,self.process))
 
 def activate(object,process,at="undefined",delay="undefined",prior=False):
     """Application function to activate passive process.""" 
+    if _e is None:
+        raise Simerror("Fatal SimPy error: simulation is not initialized (call initialize() first)")
     if not (type(process) == types.GeneratorType):
         raise Simerror("Fatal SimPy error: activating function which"+
                        " is not a generator (contains no 'yield')")
@@ -416,7 +452,10 @@ class Resource(Lister):
                 # remove from activeQ
                 self.activeQ.remove(z)
                 # put into front of waitQ
-                self.waitQ.insert(0,z)
+                self.waitQ.insert(0,z)               
+                # if self is monitored, update waitQ monitor
+                if self.monitored:
+                    self.waitMon.observe(len(self.waitQ),now())                    
                 # record that it has been preempted
                 z._preempted = 1
                 # passivate re-queued process
@@ -637,7 +676,56 @@ def holdfunc(a):
     a[0][1]._hold(a)
 
 def requestfunc(a):
-    a[0][2]._request(a)
+    """Handles 'yield request,self,res' and 'yield (request,self,res),(<code>,self,<par>)'.
+    """
+    if type(a[0][0])==tuple:
+        ## Compound yield request statement
+        ## first tuple in ((request,self,res),(xx,self,yy))
+        b=a[0][0]
+        ## b[2]==res (the resource requested)
+        ##process the first part of the compound yield statement
+        ##a[1] is the _Action instance (a[1].who==process owner, a[1].process==generator)
+        b[2]._request(arg=(b,a[1]))
+        ##deal with add-on condition to command
+        ##Trigger processes for reneging
+        class _Holder(Process):
+            """Provides timeout process"""
+            def trigger(self,delay):
+                yield hold,self,delay
+                if not proc in b[2].activeQ:
+                    reactivate(proc)
+
+        class _EventWait(Process):
+            """Provides event waiting process"""
+            def trigger(self,event):
+                yield waitevent,self,event
+                if not proc in b[2].activeQ:
+                    reactivate(proc)
+               
+        #activate it
+        proc=a[0][0][1] # the process to be woken up
+        actCode=a[0][1][0]
+        if actCode==hold:
+            proc._holder=_Holder()
+            ##                                          the timeout delay
+            activate(proc._holder,proc._holder.trigger(a[0][1][2]))
+        elif actCode==waituntil:
+            raise FatalSimerror("illegal code for reneging: waituntil")
+            #~ proc._holder=_CondWait()
+        elif actCode==waitevent:
+            proc._holder=_EventWait(proc.name)
+            ##                                          the event
+            activate(proc._holder,proc._holder.trigger(a[0][1][2]))
+        elif actCode==queueevent:
+            raise FatalSimerror("illegal code for reneging: queueevent")
+            #~ proc._holder=_QueueEvent(proc.name)
+        else:
+            raise FatalSimerror("illegal code for reneging %s"%actCode)
+        # process the first part of the yield stmnt
+        #b[2]._request((b,a[1]))
+    else:
+        ## Simple yield request command
+        a[0][2]._request(a)
 
 def releasefunc(a):
     a[0][2]._release(a)
@@ -684,14 +772,25 @@ def simulate(until=0,real_time=False,rel_speed=1):
     yield release,self,<Resource> : release 1 unit to <Resource>    
 
     yield waitevent,self,<SimEvent>|[<Evt1>,<Evt2>,<Evt3), . . . ]:
-    	wait for one or more of several events
+        wait for one or more of several events
 
 
     yield queueevent,self,<SimEvent>|[<Evt1>,<Evt2>,<Evt3), . . . ]:
-    	queue for one or more of several events
+        queue for one or more of several events
     
     yield waituntil,self,cond : wait for arbitrary condition    
 
+    EXTENSIONS:
+    Request with timeout reneging:
+    yield (request,self,<Resource>),(hold,self,<patience>) :
+        request 1 unit from <Resource>. If unit not acquired before time period
+        <patience> has passed, self leaves waitQ (reneges).
+
+    Request with event-based reneging:
+    yield (request,self,<Resource>),(waitevent,self,<event>):
+        request 1 unit from <Resource>. If <event> occurs before unit
+        acquired, leave waitQ (renege).
+
     Event notices get posted in event-list by scheduler after "yield" or by 
     "activate"/"reactivate" functions.
     
@@ -718,13 +817,27 @@ def simulate(until=0,real_time=False,rel_speed=1):
     dispatch={hold:holdfunc,request:requestfunc,release:releasefunc,
               passivate:passivatefunc,waitevent:waitevfunc,queueevent:queueevfunc,
               waituntil:waituntilfunc}
+    commandcodes=dispatch.keys()
+    commandwords={hold:"hold",request:"request",release:"release",passivate:"passivate",
+        waitevent:"waitevent",queueevent:"queueevent",waituntil:"waituntil"}
     while not _stop and _t<=_endtime:
         try:
             a=_e._nextev()        
             if not a[0]==None:
-                command = a[0][0]
+                ## 'a' is tuple "(<yield command>, <action>)"  
+                if type(a[0][0])==tuple:
+                    ##allowing for yield (request,self,res),(waituntil,self,cond)
+                    command=a[0][0][0]
+                else: 
+                    command = a[0][0]
+                if __debug__:
+                    if not command in commandcodes:
+                        raise FatalSimerror("Fatal error: illegal command: yield %s"%command)
                 dispatch[command](a)     
-        except Simerror, error:
+        except FatalSimerror,error:
+            print "SimPy: "+error.value
+            sys.exit(1)
+        except Simerror,error:
             message="SimPy: "+error.value
             _stop = True
         if _wustep:
@@ -1031,7 +1144,7 @@ def testSimEvents():
                 signal2.signal("sig 2 again")
                 print "%s: signal1, signal2 have occurred"%now()
                 yield hold,self,10
-    print"\n+++testSimEvents output"
+    print"\n\n+++testSimEvents output"
     initialize()
     signal1=SimEvent("signal 1")
     signal2=SimEvent("signal 2")
@@ -1082,7 +1195,7 @@ def testwaituntil():
                     yield release,self,item
                 yield hold,self,2 #rest
     
-    print "\n+++ nwaituntil demo output"
+    print "\n+++\nwaituntil demo output"
     initialize()
     brush=Resource(capacity=1,name="brush")
     ladder=Resource(capacity=2,name="ladder")
@@ -1100,7 +1213,6 @@ def testwaituntil():
     print simulate(until=9*60)
     
 if __name__ == "__main__":
-    print "SimPy.SimulationRT %s" %__version__
     test_demo()
     test_interrupt()
     testSimEvents()
