@@ -1,6 +1,6 @@
 #!/usr/bin/env python
-"""SimulationStep 1.5.1
-__version__ = '$Revision: 1.1.1.1.4.2 $ $Date: 2005/01/30 11:41:26 $ kgm'
+"""SimulationStep 1.6
+__version__ = '$Revision: 1.1.1.1.4.2.4.7 $ $Date: 2005/06/09 09:13:13 $ kgm'
 LICENSE:
 Copyright (C) 2002, 2005  Klaus G. Muller, Tony Vignaux
 mailto: kgmuller@xs4all.nl and Tony.Vignaux@vuw.ac.nz
@@ -33,7 +33,7 @@ Based on generators (Python 2.2 and later)
     5/9/2002: SiPy 0.9.1
     
         - Addition of '_cancel' method in class Process and supporting '_unpost' method in 
-      class __Evlist.
+          class __Evlist.
         
         - Removal of redundant 'Action' method in class Process.
         
@@ -106,7 +106,7 @@ Based on generators (Python 2.2 and later)
 
     19/11/02: Simulation version 0.6.1
         - Changed priority schemes so that higher values of Process 
-       attribute "priority" represent higher priority.
+          attribute "priority" represent higher priority.
 
     20/11/02: Simulation version 0.7
         - Major change of priority approach:
@@ -114,14 +114,14 @@ Based on generators (Python 2.2 and later)
             - Priority set by "yield request,self,res,priority"
 
             - Priority of a Process instance associated with a specific 
-          resource
+              resource
 
     25/11/02: Simulation version 0.7.1
 
         - Code cleanup and optimization
 
         - Made process attributes remainService and preempted private 
-      (_remainService and _preempted)
+         (_remainService and _preempted)
 
     11/12/2002: First process interrupt implementation
 
@@ -145,7 +145,7 @@ Based on generators (Python 2.2 and later)
         - Added 'interruptCause' as an attribute of an interrupted process
 
         - Changed definition of 'active' to 
-      'self._nextTime <> None and not self._inInterrupt'
+         'self._nextTime <> None and not self._inInterrupt'
 
         - Cleaned up test_interrupt function
 
@@ -156,7 +156,7 @@ Based on generators (Python 2.2 and later)
         - error message if no process scheduled (warning)
 
         - Ensured that upon exit from 'simulate', now() == _endtime is 
-      always valid
+          always valid
 
     2/04/2003:
 
@@ -183,18 +183,16 @@ Based on generators (Python 2.2 and later)
     
     9/6/2003: - Changed eventlist handling from pure dictionary to bisect-
                 sorted "timestamps" list of keys, resulting in greatly 
-        improved performance for models with large
+                improved performance for models with large
                 numbers of event notices with differing event times.
                 =========================================================
-        This great change was suggested by Prof. Simon Frost. 
-        Thank you, Simon! This version 1.3 is dedicated to you!
-        =========================================================
-                - Added import of Lister which supports well-structured 
-          printing of all attributes of Process and Resource instances.
+                This great change was suggested by Prof. Simon Frost. 
+                Thank you, Simon! This version 1.3 is dedicated to you!
+                =========================================================
+              - Added import of Lister which supports well-structured 
+                printing of all attributes of Process and Resource instances.
 
-    9 Nov 2003: Added capability to step through simulation event by event.
-
-    November 2003: Brought up to Simulation 1.4alpha
+    Oct 2003: Added monitored Resource instances (Monitors for activeQ and waitQ)
 
     13 Dec 2003: Merged in Monitor and Histogram
 
@@ -213,15 +211,28 @@ Based on generators (Python 2.2 and later)
     12 Jan 2005: SimPy version 1.5.1
                  Changes in this module: Monitor objects now have a default name
                                          'a_Monitor'
+                                         
+    29 Mar 2005: Start SimPy 1.6: compound "yield request" statements
+    
+    05 Jun 2005: Fixed bug in _request method -- waitMon did not work properly in
+                 preemption case
+                                         
   
 """
 from __future__ import generators
 from SimPy.Lister import *
 import bisect
 import types
+import sys
 __TESTING=False
-__version__="1.5.1 February 2005"
-if __TESTING: print "SimPy.SimulationStep %s" %__version__
+__version__="1.6 June 2005"
+if __TESTING: 
+    print "SimPy.SimulationStep %s" %__version__,
+    if __debug__:
+        print "__debug__ on"
+    else:
+        print
+
 hold=1234
 passivate=5678
 request=135
@@ -282,6 +293,11 @@ class Simerror(Exception):
 
     def __str__(self):
         return `self.value`
+        
+class FatalSimerror(Simerror):
+    def __init__(self,value):
+        Simerror.__init__(self,value)
+        self.value=value
     
 class Process(Lister):
     """Superclass of classes which may use generator functions"""
@@ -351,6 +367,21 @@ class Process(Lister):
         """
         self._inInterrupt= False
 
+    def acquired(self,res):
+        """Tests whether resource res was acquired when proces reactivated.
+        If yes, the parallel wakeup process is killed.
+        If not, process is removed from res' waitQ (reneging).
+        """
+        test=self in res.activeQ
+        
+        if test:
+            self.cancel(self._holder)
+        else:
+            res.waitQ.remove(self)
+            if res.monitored:
+                res.waitMon.observe(len(res.waitQ),t=now())
+        return test        
+
 def allEventNotices():
     """Returns string with eventlist as list of tuples (eventtime,action)"""
     return `[(_e.events[i][0],[inst.who.name for inst in _e.events[1]]) \
@@ -362,7 +393,7 @@ class __Evlist:
         #has the structure {<time1>:(ev notice1, ev notice2, . . ),
 #                           <time2>:(ev notice 3,...),..}
         self.events={}
-        #always sorted list of event times (keys for self.events)	
+        #always sorted list of event times (keys for self.events)    
         self.timestamps=[]
 
     def _post(self,what,at,prior=False):
@@ -372,7 +403,7 @@ class __Evlist:
             raise SimError("Attempt to schedule event in the past")
         if at in self.events:
             if prior:
-            #before all other event notices at this time
+                #before all other event notices at this time
                 self.events[at][0:0]= [what] 
             else:
                 self.events[at].append(what)
@@ -396,7 +427,7 @@ class __Evlist:
                     self.events[thistime].remove(n)
                     whom._nextTime = None 
                     if not self.events[thistime]:
-                    # no other ev notices for thistime
+                        # no other ev notices for thistime
                         del(self.events[thistime])
                         item_delete_point = bisect.bisect(self.timestamps,
                                               thistime)
@@ -435,10 +466,12 @@ class _Action:
         self.process=generator
 
     def __str__(self):
-        return "Action%s" %((self.who,self.process))
+        return "Action %s %s" %((self.who.name,self.process))
 
 def activate(object,process,at="undefined",delay="undefined",prior=False):
     """Application function to activate passive process.""" 
+    if _e is None:
+        raise Simerror("Fatal SimPy error: simulation is not initialized (call initialize() first)")
     if not (type(process) == types.GeneratorType):
         raise Simerror("Fatal SimPy error: activating function which"+
                        " is not a generator (contains no 'yield')")
@@ -469,7 +502,7 @@ def reactivate(object,at="undefined",delay="undefined",prior=False):
         else:
             zeit=max(_t,_t+delay)
         _e._post(_Action(who=object,generator=object._nextpoint),at=zeit,
-             prior=prior)
+                 prior=prior)
 
 
     
@@ -574,7 +607,10 @@ class Resource(Lister):
                 # remove from activeQ
                 self.activeQ.remove(z)
                 # put into front of waitQ
-                self.waitQ.insert(0,z)
+                self.waitQ.insert(0,z)               
+                # if self is monitored, update waitQ monitor
+                if self.monitored:
+                    self.waitMon.observe(len(self.waitQ),now())                    
                 # record that it has been preempted
                 z._preempted = 1
                 # passivate re-queued process
@@ -785,7 +821,7 @@ def _waitUntilFunc(proc,cond):
 
 ##end waituntil functionality
         
-def scheduler(till=0):
+def scheduler(till=0): 
     """Schedules Processes/semi-coroutines until time 'till'.
     Deprecated since version 0.5.
     """
@@ -795,7 +831,56 @@ def holdfunc(a):
     a[0][1]._hold(a)
 
 def requestfunc(a):
-    a[0][2]._request(a)
+    """Handles 'yield request,self,res' and 'yield (request,self,res),(<code>,self,<par>)'.
+    """
+    if type(a[0][0])==tuple:
+        ## Compound yield request statement
+        ## first tuple in ((request,self,res),(xx,self,yy))
+        b=a[0][0]
+        ## b[2]==res (the resource requested)
+        ##process the first part of the compound yield statement
+        ##a[1] is the _Action instance (a[1].who==process owner, a[1].process==generator)
+        b[2]._request(arg=(b,a[1]))
+        ##deal with add-on condition to command
+        ##Trigger processes for reneging
+        class _Holder(Process):
+            """Provides timeout process"""
+            def trigger(self,delay):
+                yield hold,self,delay
+                if not proc in b[2].activeQ:
+                    reactivate(proc)
+
+        class _EventWait(Process):
+            """Provides event waiting process"""
+            def trigger(self,event):
+                yield waitevent,self,event
+                if not proc in b[2].activeQ:
+                    reactivate(proc)
+               
+        #activate it
+        proc=a[0][0][1] # the process to be woken up
+        actCode=a[0][1][0]
+        if actCode==hold:
+            proc._holder=_Holder()
+            ##                                          the timeout delay
+            activate(proc._holder,proc._holder.trigger(a[0][1][2]))
+        elif actCode==waituntil:
+            raise FatalSimerror("illegal code for reneging: waituntil")
+            #~ proc._holder=_CondWait()
+        elif actCode==waitevent:
+            proc._holder=_EventWait(proc.name)
+            ##                                          the event
+            activate(proc._holder,proc._holder.trigger(a[0][1][2]))
+        elif actCode==queueevent:
+            raise FatalSimerror("illegal code for reneging: queueevent")
+            #~ proc._holder=_QueueEvent(proc.name)
+        else:
+            raise FatalSimerror("illegal code for reneging %s"%actCode)
+        # process the first part of the yield stmnt
+        #b[2]._request((b,a[1]))
+    else:
+        ## Simple yield request command
+        a[0][2]._request(a)
 
 def releasefunc(a):
     a[0][2]._release(a)
@@ -839,17 +924,28 @@ def simulate(callback=lambda :None,until=0):
 
     yield request,self,<Resource> : request 1 unit from <Resource>.
 
-    yield release,self,<Resource> : release 1 unit to <Resource>    
+    yield release,self,<Resource> : release 1 unit to <Resource>
 
     yield waitevent,self,<SimEvent>|[<Evt1>,<Evt2>,<Evt3), . . . ]:
-    	wait for one or more of several events
+        wait for one or more of several events
 
 
     yield queueevent,self,<SimEvent>|[<Evt1>,<Evt2>,<Evt3), . . . ]:
-    	queue for one or more of several events
+        queue for one or more of several events
     
     yield waituntil,self,cond : wait for arbitrary condition    
 
+    EXTENSIONS:
+    Request with timeout reneging:
+    yield (request,self,<Resource>),(hold,self,<patience>) :
+        request 1 unit from <Resource>. If unit not acquired before time period
+        <patience> has passed, self leaves waitQ (reneges).
+
+    Request with event-based reneging:
+    yield (request,self,<Resource>),(waitevent,self,<event>):
+        request 1 unit from <Resource>. If <event> occurs before unit
+        acquired, leave waitQ (renege).
+
     Event notices get posted in event-list by scheduler after "yield" or by 
     "activate"/"reactivate" functions.
 
@@ -874,14 +970,28 @@ def simulate(callback=lambda :None,until=0):
     dispatch={hold:holdfunc,request:requestfunc,release:releasefunc,
               passivate:passivatefunc,waitevent:waitevfunc,queueevent:queueevfunc,
               waituntil:waituntilfunc}
+    commandcodes=dispatch.keys()
+    commandwords={hold:"hold",request:"request",release:"release",passivate:"passivate",
+        waitevent:"waitevent",queueevent:"queueevent",waituntil:"waituntil"}
     while not _stop and _t<=_endtime:
         try:
             a=_e._nextev()        
             if not a[0]==None:
-                command = a[0][0]
+                ## 'a' is tuple "(<yield command>, <action>)"  
+                if type(a[0][0])==tuple:
+                    ##allowing for yield (request,self,res),(waituntil,self,cond)
+                    command=a[0][0][0]
+                else: 
+                    command = a[0][0]
+                if __debug__:
+                    if not command in commandcodes:
+                        raise FatalSimerror("Fatal error: illegal command: yield %s"%command)
                 dispatch[command](a)     
-        except Simerror, error:
-            message="SimPy: "+ error.value
+        except FatalSimerror,error:
+            print "SimPy: "+error.value
+            sys.exit(1)
+        except Simerror,error:
+            message="SimPy: "+error.value
             _stop = True
         if _step:
             callback()
@@ -920,8 +1030,19 @@ def simulateStep(callback=lambda :None,until=0):
         try:
             a=_e._nextev()        
             if not a[0]==None:
-                command = a[0][0]
-                dispatch[command](a)     
+                ## 'a' is tuple "(<yield command>, <action>)"  
+                if type(a[0][0])==tuple:
+                    ##allowing for yield (request,self,res),(waituntil,self,cond)
+                    command=a[0][0][0]
+                else: 
+                    command = a[0][0]
+                if __debug__:
+                    if not command in commandcodes:
+                        raise FatalSimerror("Fatal error: illegal command: yield %s"%command)
+                dispatch[command](a)         
+        except FatalSimerror,error:
+            print "SimPy: "+error.value
+            sys.exit(1)
         except Simerror, error:
             message="SimPy: "+ error.value
             _stop = True
