@@ -1,8 +1,8 @@
 #!/usr/bin/env python
-"""SimulationTrace 1.5.1
-__version__ = '$Revision: 1.1.1.3.4.2 $ $Date: 2005/01/30 11:41:26 $ kgm'
+"""SimulationTrace 1.6
+__version__ = '$Revision: 1.1.1.3.4.2.4.11 $ $Date: 2005/06/09 09:13:14 $ kgm'
 LICENSE:
-Copyright (C) 2002, 2005  Klaus G. Muller, Tony Vignaux
+Copyright (C) 2002,2005  Klaus G. Muller, Tony Vignaux
 mailto: kgmuller@xs4all.nl and Tony.Vignaux@vuw.ac.nz
 
     This library is free software; you can redistribute it and/or
@@ -60,7 +60,15 @@ Based on generators (Python 2.2 and later)
     12 Jan 2005: SimPy version 1.5.1
                  Changes in this module: Monitor objects now have a default name
                                          'a_Monitor'
-
+                                         
+    29 Mar 2005: Start SimPy 1.6: compound "yield request" statements
+    
+    05 Jun 2005: Fixed bug in _request method -- waitMon did not work properly in
+                 preemption case
+                 
+    090 Jun 2005: Added test in 'activate' to see whether 'initialize()' was called first.
+                                         
+  
 """
 from __future__ import generators
 from SimPy.Lister import *
@@ -68,8 +76,13 @@ import bisect
 import types
 import sys
 __TESTING=False
-__version__="1.5.1 February 2005"
-if __TESTING: print "SimPy.SimulationTrace %s" %__version__
+__version__="1.6 June 2005"
+if __TESTING:
+    print "SimPy.SimulationTrace %s" %__version__
+    if __debug__:
+        print "__debug__ on"
+    else:
+        print
 hold=1234
 passivate=5678
 request=135
@@ -88,7 +101,7 @@ False=0
 condQ=[]
 
 def initialize():
-    global _e,_t,_stop,trace
+    global _e,_t,_stop
     _e=__Evlist()
     _t=0
     _stop=False
@@ -117,6 +130,11 @@ class Simerror(Exception):
 
     def __str__(self):
         return `self.value`
+        
+class FatalSimerror(Simerror):
+    def __init__(self,value):
+        Simerror.__init__(self,value)
+        self.value=value
     
 class Process(Lister):
     """Superclass of classes which may use generator functions"""
@@ -124,12 +142,12 @@ class Process(Lister):
         #the reference to this Process instances single process (==generator)
         self._nextpoint=None 
         self.name=name
-        self._nextTime=None     #next activation time
+        self._nextTime=None #next activation time
         self._remainService=0
         self._preempted=0
         self._priority={}
         self._terminated=0      
-        self._inInterrupt= False    
+        self._inInterrupt= False
         self.eventsFired=[] #events waited/queued for occurred
 
     def active(self):
@@ -139,7 +157,7 @@ class Process(Lister):
         return self._nextTime == None and not self._terminated
 
     def terminated(self):
-        return self._terminated 
+        return self._terminated
 
     def interrupted(self):
         return self._inInterrupt and not self._terminated
@@ -151,7 +169,6 @@ class Process(Lister):
         """Application function to cancel all event notices for this Process instance;(should be all event
         notices for the _generator_)."""
         _e._unpost(whom=victim)
-        trace.recordCancel(self,victim)
         
     def _hold(self,a):
         if len(a[0]) == 3: 
@@ -172,7 +189,7 @@ class Process(Lister):
         # can't interrupt terminated or passive process or currently interrupted process
         if victim.active():
             save=trace._comment
-            trace._comment=None	
+            trace._comment=None    
             victim.interruptCause=self  # self causes interrupt
             left=victim._nextTime-_t
             victim.interruptLeft=left   # time left in current 'hold'
@@ -191,6 +208,21 @@ class Process(Lister):
         """
         self._inInterrupt= False
 
+    def acquired(self,res):
+        """Tests whether resource res was acquired when proces reactivated.
+        If yes, the parallel wakeup process is killed.
+        If not, process is removed from res' waitQ (reneging).
+        """
+        test=self in res.activeQ
+        
+        if test:
+            self.cancel(self._holder)
+        else:
+            res.waitQ.remove(self)
+            if res.monitored:
+                res.waitMon.observe(len(res.waitQ),t=now())
+        return test        
+
 def allEventNotices():
     """Returns string with eventlist as list of tuples (eventtime,action)"""
     return `[(_e.events[i][0],[inst.who.name for inst in _e.events[1]]) \
@@ -202,7 +234,7 @@ class __Evlist:
         #has the structure {<time1>:(ev notice1, ev notice2, . . ),
 #                           <time2>:(ev notice 3,...),..}
         self.events={}
-        #always sorted list of event times (keys for self.events)	
+        #always sorted list of event times (keys for self.events)    
         self.timestamps=[]
 
     def _post(self,what,at,prior=False):
@@ -213,7 +245,7 @@ class __Evlist:
         if at in self.events:
             if prior:
                 #before all other event notices at this time
-                self.events[at][0:0]= [what]
+                self.events[at][0:0]= [what] 
             else:
                 self.events[at].append(what)
         else: # first event notice at this time
@@ -238,8 +270,9 @@ class __Evlist:
                     if not self.events[thistime]:
                         # no other ev notices for thistime
                         del(self.events[thistime])
-                        item_delete_point = bisect.bisect(self.timestamps,thistime) 
-                        del self.timestamps[item_delete_point-1]	          
+                        item_delete_point = bisect.bisect(self.timestamps,
+                                              thistime)
+                        del self.timestamps[item_delete_point-1]              
                                               
     def _nextev(self):
         """Retrieve next event from event list"""
@@ -274,10 +307,12 @@ class _Action:
         self.process=generator
 
     def __str__(self):
-        return "Action%s" %((self.who,self.process))
+        return "Action %s %s" %((self.who.name,self.process))
 
 def activate(object,process,at="undefined",delay="undefined",prior=False):
     """Application function to activate passive process.""" 
+    if _e is None:
+        raise Simerror("Fatal SimPy error: simulation is not initialized (call initialize() first)")
     if not (type(process) == types.GeneratorType):
         raise Simerror("Fatal SimPy error: activating function which"+
                        " is not a generator (contains no 'yield')")
@@ -297,7 +332,7 @@ def reactivate(object,at="undefined",delay="undefined",prior=False):
     """Application function to reactivate a process which is active,
     suspended or passive.""" 
     # Object may be active, suspended or passive
-    if not object._terminated:
+    if not object._terminated: 
         a=Process("SimPysystem")        
         a.cancel(object)
         # object now passive
@@ -332,7 +367,7 @@ class FIFO(Queue):
     def enter(self,object):
         self.append(object)
         if self.monit:
-            self.moni.observe(len(self),t=now())        
+            self.moni.observe(len(self),t=now())
 
     def leave(self):
         a= self.pop(0)
@@ -411,7 +446,10 @@ class Resource(Lister):
                 # remove from activeQ
                 self.activeQ.remove(z)
                 # put into front of waitQ
-                self.waitQ.insert(0,z)
+                self.waitQ.insert(0,z)               
+                # if self is monitored, update waitQ monitor
+                if self.monitored:
+                    self.waitMon.observe(len(self.waitQ),now())                    
                 # record that it has been preempted
                 z._preempted = 1
                 # passivate re-queued process
@@ -419,7 +457,8 @@ class Resource(Lister):
                 # assign resource unit to preemptor
                 self.activeQ.enter(object)                    
                 # post event notice for preempting process
-                _e._post(_Action(who=object,generator=object._nextpoint),at=_t,prior=1)
+                _e._post(_Action(who=object,generator=object._nextpoint),
+                         at=_t,prior=1)
             else:
                 self.waitQ.enter(object)
                 # passivate queuing process
@@ -432,7 +471,8 @@ class Resource(Lister):
             else:
                 self.n -= 1
                 self.activeQ.enter(object)
-                _e._post(_Action(who=object,generator=object._nextpoint),at=_t,prior=1)
+                _e._post(_Action(who=object,generator=object._nextpoint),
+                at=_t,prior=1)
 
     def _release(self,arg):
         """Process release request for this resource"""
@@ -458,8 +498,8 @@ class Resource(Lister):
             # else:
             else:
                 reactivate(object,delay=0,prior=1)
-        _e._post(_Action(who=arg[1].who,generator=arg[1].who._nextpoint),at=_t,prior=1)
-
+        _e._post(_Action(who=arg[1].who,generator=arg[1].who._nextpoint),
+                at=_t,prior=1)
 
 class SimEvent(Lister):
     """Supports one-shot signalling between processes. All processes waiting for an event to occur
@@ -621,7 +661,7 @@ def _waitUntilFunc(proc,cond):
 
 ##end waituntil functionality
         
-def scheduler(till=0):
+def scheduler(till=0): 
     """Schedules Processes/semi-coroutines until time 'till'.
     Deprecated since version 0.5.
     """
@@ -631,7 +671,59 @@ def holdfunc(a):
     a[0][1]._hold(a)
 
 def requestfunc(a):
-    a[0][2]._request(a)
+    """Handles 'yield request,self,res' and 'yield (request,self,res),(<code>,self,<par>)'.
+    """
+    if type(a[0][0])==tuple:
+        ## Compound yield request statement
+        ## first tuple in ((request,self,res),(xx,self,yy))
+        b=a[0][0]
+        ## b[2]==res (the resource requested)
+        ##process the first part of the compound yield statement
+        ##a[1] is the _Action instance (a[1].who==process owner, a[1].process==generator)
+        b[2]._request(arg=(b,a[1]))
+        ##deal with add-on condition to command
+        ##Trigger processes for reneging
+        class _Holder(Process):
+            """Provides timeout process"""
+            def trigger(self,delay):
+                yield hold,self,delay
+                if not proc in b[2].activeQ:
+                    reactivate(proc)
+
+        class _EventWait(Process):
+            """Provides event waiting process"""
+            def trigger(self,event):
+                yield waitevent,self,event
+                if not proc in b[2].activeQ:
+                    reactivate(proc)
+
+        #activate it
+        proc=a[0][0][1] # the process to be woken up
+        actCode=a[0][1][0]
+        trace.tstop()
+        if actCode==hold:
+            proc._holder=_Holder(name="RENEGE-hold for "+proc.name)
+            ##                                          the timeout delay
+            activate(proc._holder,proc._holder.trigger(a[0][1][2]))
+        elif actCode==waituntil:
+            raise FatalSimerror("illegal code for reneging: waituntil")
+            #~ proc._holder=_CondWait()
+        elif actCode==waitevent:
+            proc._holder=_EventWait(name="RENEGE-waitevent for "+proc.name)
+            ##                                          the event
+            activate(proc._holder,proc._holder.trigger(a[0][1][2]))
+        elif actCode==queueevent:
+            raise FatalSimerror("illegal code for reneging: queueevent")
+            #~ proc._holder=_QueueEvent(proc.name)
+        else:
+            raise FatalSimerror("illegal code for reneging %s"%actCode)
+        #activate(proc._holder,proc._holder.trigger(a[0][1][2]),delay=0.00000001)
+        # process the first part of the yield stmnt
+        #b[2]._request((b,a[1]))
+        trace.tstart()
+    else:
+        ## Simple yield request command
+        a[0][2]._request(a)
 
 def releasefunc(a):
     a[0][2]._release(a)
@@ -676,17 +768,28 @@ def simulate(until=0):
 
     yield request,self,<Resource> : request 1 unit from <Resource>.
 
-    yield release,self,<Resource> : release 1 unit to <Resource>    
+    yield release,self,<Resource> : release 1 unit to <Resource>
 
     yield waitevent,self,<SimEvent>|[<Evt1>,<Evt2>,<Evt3), . . . ]:
-    	wait for one or more of several events
-    	
+        wait for one or more of several events
+        
 
     yield queueevent,self,<SimEvent>|[<Evt1>,<Evt2>,<Evt3), . . . ]:
-    	queue for one or more of several events
+        queue for one or more of several events
     
     yield waituntil,self,cond : wait for arbitrary condition    
 
+    EXTENSIONS:
+    Request with timeout reneging:
+    yield (request,self,<Resource>),(hold,self,<patience>) :
+        request 1 unit from <Resource>. If unit not acquired before time period
+        <patience> has passed, self leaves waitQ (reneges).
+
+    Request with event-based reneging:
+    yield (request,self,<Resource>),(waitevent,self,<event>):
+        request 1 unit from <Resource>. If <event> occurs before unit
+        acquired, leave waitQ (renege).
+
     Event notices get posted in event-list by scheduler after "yield" or by 
     "activate"/"reactivate" functions.
     
@@ -704,16 +807,30 @@ def simulate(until=0):
     dispatch={hold:holdfunc,request:requestfunc,release:releasefunc,
               passivate:passivatefunc,waitevent:waitevfunc,queueevent:queueevfunc,
               waituntil:waituntilfunc}
+    commandcodes=dispatch.keys()
+    commandwords={hold:"hold",request:"request",release:"release",passivate:"passivate",
+        waitevent:"waitevent",queueevent:"queueevent",waituntil:"waituntil"}
     while not _stop and _t<=_endtime:
         try:
-            a=_e._nextev()        
+            a=_e._nextev()          
             if not a[0]==None:
-                command = a[0][0]
+                ## 'a' is tuple "(<yield command>, <action>)"
+                if type(a[0][0])==tuple:
+                    ##allowing for yield (request,self,res),(waituntil,self,cond)
+                    command=a[0][0][0]
+                else:
+                    command = a[0][0]
+                if __debug__:
+                    if not command in commandcodes:
+                        raise FatalSimerror("Fatal error: illegal command: yield %s"%command)
                 dispatch[command](a)
                 trace.recordEvent(command,a)
             else:
                 if not a==(None,): #not at endtime!
                     trace.tterminated(a[1])
+        except FatalSimerror,error:
+            print "SimPy: "+error.value
+            sys.exit(1)
         except Simerror, error:
             message="SimPy: "+error.value
             _stop = True 
@@ -840,10 +957,22 @@ class Trace(Lister):
     
     def recordEvent(self,command,whole):
         if self.ifTrace(Trace.commands[command] in self.toTrace):
-            print >>self.outfile, now(),Trace.commands[command],"<"+whole[0][1].name+">",\
-                  Trace.commandsproc[command](whole)
-            if self._comment:
-                print >>self.outfile,"----",self._comment
+            if not type(whole[0][0])==tuple:
+                print >>self.outfile, now(),Trace.commands[command],"<"+whole[0][1].name+">",\
+                      Trace.commandsproc[command](whole)
+                if self._comment:
+                    print >>self.outfile,"----",self._comment
+            else:
+                ##print >>self.outfile, "[WHOLE]",whole,"\n[END WHOLE]"
+                print >>self.outfile, now(),Trace.commands[command],"<"+whole[0][0][1].name+">"+\
+                      Trace.commandsproc[command](whole[0])
+                print >>self.outfile,"|| RENEGE COMMAND:"
+                command1=whole[0][1][0]
+                print >>self.outfile,"||\t",Trace.commands[command1],"<"+whole[0][1][1].name+">",\
+                      Trace.commandsproc[command1]((whole[0][1],))
+                if self._comment:
+                    print >>self.outfile,"----",self._comment
+                
         self._comment=None
 
     def recordInterrupt(self,who,victim):
@@ -1258,15 +1387,214 @@ def testwaituntil():
         print "%s needs %s for his job" %(who.name,[x.name for x in who.heNeeds])
     print
     print simulate(until=9*60)
+    
+## -------------------------------------------------------------
+##                    TEST COMPOUND "YIELD REQUEST" COMMANDS
+## -------------------------------------------------------------
+
+## -------------------------------------------------------------
+##             TEST "yield (request,self,res),(hold,self,delay)"
+##                   == timeout renege
+## -------------------------------------------------------------
+
+class JobTO(Process):
+   """ Job class for testing timeout reneging
+   """
+   def __init__(self,server=None,name=""):
+        Process.__init__(self,name)
+        self.res=server
+        self.gotResource=None
+        
+   def execute(self,timeout,usetime):       
+        yield (request,self,self.res),(hold,self,timeout)
+        if self.acquired(self.res):
+            self.gotResource=True
+            yield hold,self,usetime
+            yield release,self,self.res
+        else:
+            self.gotResource=False
+            
+
+def testNoTimeout():
+    """Test that resource gets acquired without timeout
+    """
+    res=Resource(name="Server",capacity=1)
+    initialize()
+    usetime=5
+    timeout=1000000
+    j1=JobTO(server=res,name="Job_1")
+    activate(j1,j1.execute(timeout=timeout,usetime=usetime))
+    j2=JobTO(server=res,name="Job_2")
+    activate(j2,j2.execute(timeout=timeout,usetime=usetime))      
+    simulate(until=2*usetime)
+    assert now()==2*usetime,"time not ==2*usetime" 
+    assert j1.gotResource and j2.gotResource,\
+        "at least one job failed to get resource"
+    assert not (res.waitQ or res.activeQ),\
+        "job waiting or using resource"
+            
+def testTimeout1():
+    """Test that timeout occurs when resource busy
+    """
+    res=Resource(name="Server",capacity=1)
+    initialize()
+    usetime=5
+    timeout=3
+    j1=JobTO(server=res,name="Job_1")
+    activate(j1,j1.execute(timeout=timeout,usetime=usetime))
+    j2=JobTO(server=res,name="Job_2")
+    activate(j2,j2.execute(timeout=timeout,usetime=usetime))      
+    simulate(until=2*usetime)
+    assert(now()==usetime),"time not ==usetime"
+    assert(j1.gotResource),"Job_1 did not get resource"
+    assert(not j2.gotResource),"Job_2 did not renege"
+    assert not (res.waitQ or res.activeQ),\
+        "job waiting or using resource"
+
+def testTimeout2():
+    """Test that timeout occurs when resource has no capacity free
+    """
+    res=Resource(name="Server",capacity=0)
+    initialize()
+    usetime=5
+    timeout=3
+    j1=JobTO(server=res,name="Job_1")
+    activate(j1,j1.execute(timeout=timeout,usetime=usetime))
+    j2=JobTO(server=res,name="Job_2")
+    activate(j2,j2.execute(timeout=timeout,usetime=usetime))      
+    simulate(until=2*usetime)
+    assert now()==timeout,"time %s not == timeout"%now()
+    assert not j1.gotResource,"Job_1 got resource"
+    assert not j2.gotResource,"Job_2 got resource"
+    assert not (res.waitQ or res.activeQ),\
+        "job waiting or using resource"  
+
+## ------------------------------------------------------------------
+##             TEST "yield (request,self,res),(waitevent,self,event)"
+##                   == event renege
+## ------------------------------------------------------------------
+class JobEvt(Process):
+   """ Job class for testing event reneging
+   """
+   def __init__(self,server=None,name=""):
+        Process.__init__(self,name)
+        self.res=server
+        self.gotResource=None
+        
+   def execute(self,event,usetime):       
+        yield (request,self,self.res),(waitevent,self,event)
+        if self.acquired(self.res):
+            self.gotResource=True
+            yield hold,self,usetime
+            yield release,self,self.res
+        else:
+            self.gotResource=False
+            
+class JobEvtMulti(Process):
+   """ Job class for testing event reneging with multi-event lists
+   """
+   def __init__(self,server=None,name=""):
+        Process.__init__(self,name)
+        self.res=server
+        self.gotResource=None
+        
+   def execute(self,eventlist,usetime):       
+        yield (request,self,self.res),(waitevent,self,eventlist)
+        if self.acquired(self.res):
+            self.gotResource=True
+            yield hold,self,usetime
+            yield release,self,self.res
+        else:
+            self.gotResource=False
+            
+class FireEvent(Process):
+    """Fires reneging event
+    """
+    def fire(self,fireDelay,event):
+        yield hold,self,fireDelay
+        event.signal()
+            
+def testNoEvent():
+    """Test that processes acquire resource normally if no event fires
+    """
+    res=Resource(name="Server",capacity=1)
+    event=SimEvent("Renege_trigger") #never gets fired
+    initialize()
+    usetime=5
+    j1=JobEvt(server=res,name="Job_1")
+    activate(j1,j1.execute(event=event,usetime=usetime))
+    j2=JobEvt(server=res,name="Job_2")
+    activate(j2,j2.execute(event=event,usetime=usetime))      
+    simulate(until=2*usetime)
+    # Both jobs should get server (in sequence)
+    assert now()==2*usetime,"time not ==2*usetime" 
+    assert j1.gotResource and j2.gotResource,\
+        "at least one job failed to get resource"
+    assert not (res.waitQ or res.activeQ),\
+        "job waiting or using resource"
+ 
+def testWaitEvent1():
+    """Test that signalled event leads to renege when resource busy
+    """
+    res=Resource(name="Server",capacity=1)
+    initialize()
+    event=SimEvent("Renege_trigger")
+    usetime=5
+    eventtime=1
+    j1=JobEvt(server=res,name="Job_1")
+    activate(j1,j1.execute(event=event,usetime=usetime))
+    j2=JobEvt(server=res,name="Job_2")
+    activate(j2,j2.execute(event=event,usetime=usetime))
+    f=FireEvent(name="FireEvent")
+    activate(f,f.fire(fireDelay=eventtime,event=event))           
+    simulate(until=2*usetime)
+    # Job_1 should get server, Job_2 renege
+    assert(now()==usetime),"time not ==usetime"
+    assert(j1.gotResource),"Job_1 did not get resource"
+    assert(not j2.gotResource),"Job_2 did not renege"
+    assert not (res.waitQ or res.activeQ),\
+        "job waiting or using resource"
+            
+def testWaitEvent2():
+    """Test that renege-triggering event can be one of an event list
+    """
+    res=Resource(name="Server",capacity=1)
+    initialize()
+    event1=SimEvent("Renege_trigger_1")
+    event2=SimEvent("Renege_trigger_2")
+    usetime=5
+    eventtime=1 #for both events
+    j1=JobEvtMulti(server=res,name="Job_1")
+    activate(j1,j1.execute(eventlist=[event1,event2],usetime=usetime))
+    j2=JobEvtMulti(server=res,name="Job_2")
+    activate(j2,j2.execute(eventlist=[event1,event2],usetime=usetime))
+    f1=FireEvent(name="FireEvent_1")
+    activate(f1,f1.fire(fireDelay=eventtime,event=event1))
+    f2=FireEvent(name="FireEvent_2")
+    activate(f2,f2.fire(fireDelay=eventtime,event=event2))             
+    simulate(until=2*usetime)
+    # Job_1 should get server, Job_2 should renege
+    assert(now()==usetime),"time not ==usetime"
+    assert(j1.gotResource),"Job_1 did not get resource"
+    assert(not j2.gotResource),"Job_2 did not renege"
+    assert not (res.waitQ or res.activeQ),\
+        "job waiting or using resource"
 
 if __name__ == "__main__":
-    print "SimPy.SimulationTrace %s" %__version__
     trace=Trace()
+    testNoTimeout()
+    testTimeout1()
+    testTimeout2()
+    testNoEvent()
+    testWaitEvent1()
+    testWaitEvent2()
     trace=Trace(end=4000)
     test_demo()
     trace=Trace(end=2000)
     test_interrupt()
     testSimEvents()
     testwaituntil()
+    
+    
 else:
     trace=Trace()    
