@@ -1,7 +1,7 @@
 #!/usr/bin/env python
 # testSimPyStep.py test routines
-#        $Author: kgmuller $              $Revision: 1.1.1.1.4.1 $
-#        $Date: 2005/01/30 11:41:27 $
+#        $Author: kgmuller $              $Revision: 1.1.1.1.4.1.4.4 $
+#        $Date: 2005/06/06 04:32:21 $
 # 2002 11 15 Added tests for priority queues and preemption
 # 2002 11 22 testing problem in accum
 # 2003 03 30 added tests for SEP001v17 interrupts
@@ -14,12 +14,13 @@
 # 2004 05 15 first version of testSimPyStep; just
 #            tests compatibility with Simulation.py
 # 2004 09 17 added tests for waitevent, queueevent, waituntil (new in 1.5)
-# '$Revision: 1.1.1.1.4.1 $ $Date: 2005/01/30 11:41:27 $ kgm'
+# 2005 05 19 added tests for compound yield statements (reneging)
+# '$Revision: 1.1.1.1.4.1.4.4 $ $Date: 2005/06/06 04:32:21 $ kgm'
 
 """Testing SimulationStep compatibility with Simulation
 """
-__version__ = '1.5.1 February 2005'
-Revision = '$Revision: 1.1.1.1.4.1 $'
+__version__ = '1.6 June 2005'
+Revision = '$Revision: 1.1.1.1.4.1.4.4 $'
 
 from __future__ import generators
 from SimPy.MonitorTest import *
@@ -908,12 +909,417 @@ def makeWSuite():
    testwaituntil1=makeWtestcase("testwaituntil1")
    suite.addTests([testwaituntil1])
    return suite
+   
+## -------------------------------------------------------------
+##                    TEST COMPOUND "YIELD REQUEST" COMMANDS
+## -------------------------------------------------------------
+
+## -------------------------------------------------------------
+##             TEST "yield (request,self,res),(hold,self,delay)"
+##                   == timeout renege
+##             for both unmonitored and monitored resources
+## -------------------------------------------------------------
+
+class JobTO(Process):
+   """ Job class for testing timeout reneging
+   """
+   def __init__(self,server=None,name=""):
+        Process.__init__(self,name)
+        self.res=server
+        self.gotResource=None
+        
+   def execute(self,timeout,usetime):       
+        yield (request,self,self.res),(hold,self,timeout)
+        if self.acquired(self.res):
+            self.gotResource=True
+            yield hold,self,usetime
+            yield release,self,self.res
+        else:
+            self.gotResource=False
+            
+class JobTO_P(Process):
+   """ Job class for testing timeout reneging with priorities
+   """
+   def __init__(self,server=None,name=""):
+        Process.__init__(self,name)
+        self.res=server
+        self.gotResource=None
+        
+   def execute(self,timeout,usetime,priority):       
+        yield (request,self,self.res,priority),(hold,self,timeout)
+        if self.acquired(self.res):
+            self.gotResource=True
+            yield hold,self,usetime
+            yield release,self,self.res
+        else:
+            self.gotResource=False
+            
+class makeTimeoutTestcase(unittest.TestCase):
+    """ Tests of "yield (request,self,res),(hold,self,delay)"
+        timeout reneging command
+    """
+    def testNoTimeout(self):
+        """Test that resource gets acquired without timeout
+        """
+        res=Resource(name="Server",capacity=1)
+        initialize()
+        usetime=5
+        timeout=1000000
+        j1=JobTO(server=res,name="Job_1")
+        activate(j1,j1.execute(timeout=timeout,usetime=usetime))
+        j2=JobTO(server=res,name="Job_2")
+        activate(j2,j2.execute(timeout=timeout,usetime=usetime))      
+        simulate(until=2*usetime)
+        assert now()==2*usetime,"time not ==2*usetime" 
+        assert j1.gotResource and j2.gotResource,\
+            "at least one job failed to get resource"
+        assert not (res.waitQ or res.activeQ),\
+            "job waiting or using resource"
+            
+    def testNoTimeoutM(self):
+        """Test that resource gets acquired without timeout.
+           Resource monitored.
+        """
+        res=Resource(name="Server",capacity=1,monitored=True)
+        initialize()
+        usetime=5
+        timeout=1000000
+        j1=JobTO(server=res,name="Job_1")
+        activate(j1,j1.execute(timeout=timeout,usetime=usetime))
+        j2=JobTO(server=res,name="Job_2")
+        activate(j2,j2.execute(timeout=timeout,usetime=usetime))      
+        simulate(until=2*usetime)
+        assert now()==2*usetime,"time not ==2*usetime" 
+        assert j1.gotResource and j2.gotResource,\
+            "at least one job failed to get resource"
+        assert not (res.waitQ or res.activeQ),\
+            "job waiting or using resource"
+        assert res.waitMon==[[0,1],[usetime,0]],"res.waitMon wrong: %s"%res.waitMon
+            
+    def testTimeout1(self):
+        """Test that timeout occurs when resource busy
+        """
+        res=Resource(name="Server",capacity=1)
+        initialize()
+        usetime=5
+        timeout=3
+        j1=JobTO(server=res,name="Job_1")
+        activate(j1,j1.execute(timeout=timeout,usetime=usetime))
+        j2=JobTO(server=res,name="Job_2")
+        activate(j2,j2.execute(timeout=timeout,usetime=usetime))      
+        simulate(until=2*usetime)
+        assert(now()==usetime),"time not ==usetime"
+        assert(j1.gotResource),"Job_1 did not get resource"
+        assert(not j2.gotResource),"Job_2 did not renege"
+        assert not (res.waitQ or res.activeQ),\
+            "job waiting or using resource"
+            
+    def testTimeout1M(self):
+        """Test that timeout occurs when resource busy.
+           Resource monitored.
+        """
+        res=Resource(name="Server",capacity=1,monitored=True)
+        initialize()
+        usetime=5
+        timeout=3
+        j1=JobTO(server=res,name="Job_1")
+        activate(j1,j1.execute(timeout=timeout,usetime=usetime))
+        j2=JobTO(server=res,name="Job_2")
+        activate(j2,j2.execute(timeout=timeout,usetime=usetime))      
+        simulate(until=2*usetime)
+        assert(now()==usetime),"time not == usetime"
+        assert(j1.gotResource),"Job_1 did not get resource"
+        assert(not j2.gotResource),"Job_2 did not renege"
+        assert not (res.waitQ or res.activeQ),\
+            "job waiting or using resource"
+        assert res.waitMon==[[0,1],[timeout,0]],"res.waitMon wrong: %s"%res.waitMon
+        
+    def testTimeout_MP(self):
+        """Test that timeout occurs when resource busy.
+           Resource monitored. Requests with priority and preemption.
+        """
+        res=Resource(name="Server",capacity=1,monitored=True,qType=PriorityQ,preemptable=True)
+        initialize()
+        usetime=5
+        timeout=3
+        j1=JobTO_P(server=res,name="Job_1")
+        activate(j1,j1.execute(timeout=timeout,usetime=usetime,priority=1))
+        j2=JobTO_P(server=res,name="Job_2")
+        j2_arrival=1
+        activate(j2,j2.execute(timeout=timeout,usetime=usetime,priority=5),at=j2_arrival)
+        j3=JobTO_P(server=res,name="Job_2")
+        j3_arrival=2
+        activate(j3,j3.execute(timeout=timeout,usetime=usetime,priority=10),at=j3_arrival)           
+        simulate(until=3*usetime)
+        assert(now()== 3*usetime),"time not == 2* usetime, but %s"%now()
+        assert(j1.gotResource),"Job_1 did not get resource"
+        assert(j2.gotResource),"Job_2 did renege"
+        assert(j2.gotResource),"Job_3 did renege"
+        assert not (res.waitQ or res.activeQ),\
+            "job waiting or using resource"
+        assert res.waitMon==[[j2_arrival,1],[j3_arrival,2],[usetime+j3_arrival,1],[usetime+j2_arrival+usetime,0]],\
+             "res.waitMon wrong: %s"%res.waitMon
+
+    def testTimeout2(self):
+        """Test that timeout occurs when resource has no capacity free
+        """
+        res=Resource(name="Server",capacity=0)
+        initialize()
+        usetime=5
+        timeout=3
+        j1=JobTO(server=res,name="Job_1")
+        activate(j1,j1.execute(timeout=timeout,usetime=usetime))
+        j2=JobTO(server=res,name="Job_2")
+        activate(j2,j2.execute(timeout=timeout,usetime=usetime))      
+        simulate(until=2*usetime)
+        assert now()==timeout,"time %s not == timeout"%now()
+        assert not j1.gotResource,"Job_1 got resource"
+        assert not j2.gotResource,"Job_2 got resource"
+        assert not (res.waitQ or res.activeQ),\
+            "job waiting or using resource"  
+            
+    def testTimeout2M(self):
+        """Test that timeout occurs when resource has no capacity free.
+           Resource monitored.
+        """
+        res=Resource(name="Server",capacity=0,monitored=True)
+        initialize()
+        usetime=5
+        timeout=3
+        j1=JobTO(server=res,name="Job_1")
+        activate(j1,j1.execute(timeout=timeout,usetime=usetime))
+        j2=JobTO(server=res,name="Job_2")
+        activate(j2,j2.execute(timeout=timeout,usetime=usetime))      
+        simulate(until=2*usetime)
+        assert now()==timeout,"time %s not == timeout"%now()
+        assert not j1.gotResource,"Job_1 got resource"
+        assert not j2.gotResource,"Job_2 got resource"
+        assert not (res.waitQ or res.activeQ),\
+            "job waiting or using resource"  
+        assert res.waitMon==[[0,1],[0,2],[timeout,1],[timeout,0]],\
+            "res.waitMon is wrong: %s"%res.waitMon
+            
+def makeTOSuite():
+    suite = unittest.TestSuite()
+    testNoTimeout = makeTimeoutTestcase("testNoTimeout")
+    testNoTimeoutM = makeTimeoutTestcase("testNoTimeoutM")
+    testTimeout1=makeTimeoutTestcase("testTimeout1")
+    testTimeout1M=makeTimeoutTestcase("testTimeout1M")
+    testTimeout_MP=makeTimeoutTestcase("testTimeout_MP")       
+    testTimeout2=makeTimeoutTestcase("testTimeout2")
+    testTimeout2M=makeTimeoutTestcase("testTimeout2M")
+    suite.addTests([testNoTimeout,testNoTimeoutM,
+                    testTimeout1,testTimeout1M,testTimeout_MP,
+                   testTimeout2,testTimeout2M]) 
+    return suite
+
+## ------------------------------------------------------------------
+##             TEST "yield (request,self,res),(waitevent,self,event)"
+##                   == event renege
+##             for both unmonitored and monitored resources
+## ------------------------------------------------------------------
+class JobEvt(Process):
+   """ Job class for testing event reneging
+   """
+   def __init__(self,server=None,name=""):
+        Process.__init__(self,name)
+        self.res=server
+        self.gotResource=None
+        
+   def execute(self,event,usetime):       
+        yield (request,self,self.res),(waitevent,self,event)
+        if self.acquired(self.res):
+            self.gotResource=True
+            yield hold,self,usetime
+            yield release,self,self.res
+        else:
+            self.gotResource=False
+            
+class JobEvtMulti(Process):
+   """ Job class for testing event reneging with multi-event lists
+   """
+   def __init__(self,server=None,name=""):
+        Process.__init__(self,name)
+        self.res=server
+        self.gotResource=None
+        
+   def execute(self,eventlist,usetime):       
+        yield (request,self,self.res),(waitevent,self,eventlist)
+        if self.acquired(self.res):
+            self.gotResource=True
+            yield hold,self,usetime
+            yield release,self,self.res
+        else:
+            self.gotResource=False
+            
+class FireEvent(Process):
+    """Fires reneging event
+    """
+    def fire(self,fireDelay,event):
+        yield hold,self,fireDelay
+        event.signal()
+            
+class makeEventRenegeTestcase(unittest.TestCase):
+    """Tests of "yield (request,self,res),(waiteevent,self,event)" 
+       event reneging command 
+    """
+    def testNoEvent(self):
+        """Test that processes acquire resource normally if no event fires
+        """
+        res=Resource(name="Server",capacity=1)
+        event=SimEvent("Renege_trigger") #never gets fired
+        initialize()
+        usetime=5
+        j1=JobEvt(server=res,name="Job_1")
+        activate(j1,j1.execute(event=event,usetime=usetime))
+        j2=JobEvt(server=res,name="Job_2")
+        activate(j2,j2.execute(event=event,usetime=usetime))      
+        simulate(until=2*usetime)
+        # Both jobs should get server (in sequence)
+        assert now()==2*usetime,"time not ==2*usetime" 
+        assert j1.gotResource and j2.gotResource,\
+            "at least one job failed to get resource"
+        assert not (res.waitQ or res.activeQ),\
+            "job waiting or using resource"
+ 
+    def testNoEventM(self):
+        """Test that processes acquire resource normally if no event fires.
+           Resource monitored.
+        """
+        res=Resource(name="Server",capacity=1,monitored=True)
+        event=SimEvent("Renege_trigger") #never gets fired
+        initialize()
+        usetime=5
+        j1=JobEvt(server=res,name="Job_1")
+        activate(j1,j1.execute(event=event,usetime=usetime))
+        j2=JobEvt(server=res,name="Job_2")
+        activate(j2,j2.execute(event=event,usetime=usetime))      
+        simulate(until=2*usetime)
+        # Both jobs should get server (in sequence)
+        assert now()==2*usetime,"time not ==2*usetime" 
+        assert j1.gotResource and j2.gotResource,\
+            "at least one job failed to get resource"
+        assert not (res.waitQ or res.activeQ),\
+            "job waiting or using resource" 
+        assert res.waitMon==[[0,1],[usetime,0]],"res.waitMoni is wrong: %s"%res.waitMon
+            
+    def testWaitEvent1(self):
+        """Test that signalled event leads to renege when resource busy
+        """
+        res=Resource(name="Server",capacity=1)
+        initialize()
+        event=SimEvent("Renege_trigger")
+        usetime=5
+        eventtime=1
+        j1=JobEvt(server=res,name="Job_1")
+        activate(j1,j1.execute(event=event,usetime=usetime))
+        j2=JobEvt(server=res,name="Job_2")
+        activate(j2,j2.execute(event=event,usetime=usetime))
+        f=FireEvent(name="FireEvent")
+        activate(f,f.fire(fireDelay=eventtime,event=event))           
+        simulate(until=2*usetime)
+        # Job_1 should get server, Job_2 renege
+        assert(now()==usetime),"time not ==usetime"
+        assert(j1.gotResource),"Job_1 did not get resource"
+        assert(not j2.gotResource),"Job_2 did not renege"
+        assert not (res.waitQ or res.activeQ),\
+            "job waiting or using resource"
+            
+    def testWaitEvent1M(self):
+        """Test that signalled event leads to renege when resource busy.
+           Resource monitored.
+        """
+        res=Resource(name="Server",capacity=1,monitored=True)
+        initialize()
+        event=SimEvent("Renege_trigger")
+        usetime=5
+        eventtime=1
+        j1=JobEvt(server=res,name="Job_1")
+        activate(j1,j1.execute(event=event,usetime=usetime))
+        j2=JobEvt(server=res,name="Job_2")
+        activate(j2,j2.execute(event=event,usetime=usetime))
+        f=FireEvent(name="FireEvent")
+        activate(f,f.fire(fireDelay=eventtime,event=event))           
+        simulate(until=2*usetime)
+        # Job_1 should get server, Job_2 renege
+        assert(now()==usetime),"time not ==usetime"
+        assert(j1.gotResource),"Job_1 did not get resource"
+        assert(not j2.gotResource),"Job_2 did not renege"
+        assert not (res.waitQ or res.activeQ),\
+            "job waiting or using resource"
+        assert res.waitMon==[[0,1],[eventtime,0]],"res.waitMon is wrong: %s"%res.waitMon
+            
+    def testWaitEvent2(self):
+        """Test that renege-triggering event can be one of an event list
+        """
+        res=Resource(name="Server",capacity=1)
+        initialize()
+        event1=SimEvent("Renege_trigger_1")
+        event2=SimEvent("Renege_trigger_2")
+        usetime=5
+        eventtime=1 #for both events
+        j1=JobEvtMulti(server=res,name="Job_1")
+        activate(j1,j1.execute(eventlist=[event1,event2],usetime=usetime))
+        j2=JobEvtMulti(server=res,name="Job_2")
+        activate(j2,j2.execute(eventlist=[event1,event2],usetime=usetime))
+        f1=FireEvent(name="FireEvent_1")
+        activate(f1,f1.fire(fireDelay=eventtime,event=event1))
+        f2=FireEvent(name="FireEvent_2")
+        activate(f2,f2.fire(fireDelay=eventtime,event=event2))             
+        simulate(until=2*usetime)
+        # Job_1 should get server, Job_2 should renege
+        assert(now()==usetime),"time not ==usetime"
+        assert(j1.gotResource),"Job_1 did not get resource"
+        assert(not j2.gotResource),"Job_2 did not renege"
+        assert not (res.waitQ or res.activeQ),\
+            "job waiting or using resource"
+            
+    def testWaitEvent2M(self):
+        """Test that renege-triggering event can be one of an event list.
+           Resource monitored.
+        """
+        res=Resource(name="Server",capacity=1,monitored=True)
+        initialize()
+        event1=SimEvent("Renege_trigger_1")
+        event2=SimEvent("Renege_trigger_2")
+        usetime=5
+        eventtime=1 #for both events
+        j1=JobEvtMulti(server=res,name="Job_1")
+        activate(j1,j1.execute(eventlist=[event1,event2],usetime=usetime))
+        j2=JobEvtMulti(server=res,name="Job_2")
+        activate(j2,j2.execute(eventlist=[event1,event2],usetime=usetime))
+        f1=FireEvent(name="FireEvent_1")
+        activate(f1,f1.fire(fireDelay=eventtime,event=event1))
+        f2=FireEvent(name="FireEvent_2")
+        activate(f2,f2.fire(fireDelay=eventtime,event=event2))             
+        simulate(until=2*usetime)
+        # Job_1 should get server, Job_2 should renege
+        assert(now()==usetime),"time not ==usetime"
+        assert(j1.gotResource),"Job_1 did not get resource"
+        assert(not j2.gotResource),"Job_2 did not renege"
+        assert not (res.waitQ or res.activeQ),\
+            "job waiting or using resource"
+        assert res.waitMon==[[0,1],[eventtime,0]],"res.waitMon is wrong: %s"%res.waitMon
+    
+def makeEvtRenegeSuite():
+    suite = unittest.TestSuite()
+    testNoEvent = makeEventRenegeTestcase("testNoEvent")
+    testNoEventM = makeEventRenegeTestcase("testNoEventM")
+    testWaitEvent1=makeEventRenegeTestcase("testWaitEvent1")
+    testWaitEvent1M=makeEventRenegeTestcase("testWaitEvent1M")    
+    testWaitEvent2=makeEventRenegeTestcase("testWaitEvent2")
+    testWaitEvent2M=makeEventRenegeTestcase("testWaitEvent2M")
+    
+    suite.addTests([testNoEvent,testNoEventM,testWaitEvent1,testWaitEvent1M,
+                    testWaitEvent2,testWaitEvent2M]) 
+    return suite
 
 if __name__ == '__main__':
     alltests = unittest.TestSuite((makeSSuite(),makeRSuite(),
                                    makeMSuite(),makeHSuite(),
                                    makeISuite(),makePSuite(),
-                                   makeESuite(),makeWSuite()))
+                                   makeESuite(),makeWSuite(),
+                                   makeTOSuite(),makeEvtRenegeSuite()))
     runner = unittest.TextTestRunner()
     runner.run(alltests)
 
